@@ -5,7 +5,7 @@ from __future__ import annotations
 from abc import ABC, abstractmethod
 from collections.abc import Iterator
 from dataclasses import dataclass
-from functools import cached_property
+from functools import lru_cache
 from typing import Any, Callable, ClassVar, Optional, Type, Union
 
 import h5py
@@ -20,6 +20,7 @@ from .data_management import (
     SavableData,
     phase_unwrapping,
 )
+from .multibanding import reduced_frequency_array
 
 # from .downsampling_interpolation import DownsamplingTraining
 from .taylorf2 import (
@@ -52,10 +53,8 @@ class WaveformGenerator(ABC):
 
     Users may wish to leave the Post-Newtonian model already implemented here
     and only switch TEOBResumS to another waveform template:
-    the easiest way to accomplish this is to subclass TEOBResumSGenerator
+    the easiest way to accomplish this is to subclass :class:`BarePostNewtonianGenerator`
     and only override its :meth:`effective_one_body_waveform` method.
-
-
     """
 
     @abstractmethod
@@ -120,14 +119,19 @@ class WaveformGenerator(ABC):
 
     @abstractmethod
     def effective_one_body_waveform(
-        self, params: "WaveformParameters"
+        self, params: "WaveformParameters", frequencies: Optional[list[float]] = None
     ) -> tuple[np.ndarray, np.ndarray]:
-        r"""Waveform
+        r"""Waveform computed according to the comparatively slower
+        effective-one-body method.
 
         Parameters
         ----------
         params : WaveformParameters
                 Parameters of the binary system for which to generate the waveform.
+        frequencies : list[float], optional
+                Frequencies at which to compute the waveform, in natural units.
+                Defaults to None, which means the EOB generator will choose
+                the frequencies at which to compute the waveform.
 
         Returns
         -------
@@ -143,6 +147,7 @@ class WaveformGenerator(ABC):
     def generate_residuals(
         self,
         params: "WaveformParameters",
+        frequencies: Optional[list[float]] = None,
         downsampling_indices: Optional[DownsamplingIndices] = None,
     ) -> tuple[np.ndarray, np.ndarray]:
         """Compute the residuals of the :func:`effective_one_body_waveform`
@@ -156,6 +161,14 @@ class WaveformGenerator(ABC):
         ----------
         params : WaveformParameters
                 Parameters for which to compute the residuals.
+        frequencies : list[float], optional
+                Frequencies at which to compute the residuals,
+                in natural units.
+                If this parameter is given, the downsampling_indices
+                should index this frequency array.
+                Defaults to None, meaning that the frequencies computed
+                by the :meth:`effective_one_body_waveform` method
+                are used.
         downsampling_indices : Optional[DownsamplingIndices]
                 Indices at which to compute the residuals.
                 If not provided (default) the waveform is given at
@@ -169,7 +182,7 @@ class WaveformGenerator(ABC):
         (
             frequencies_eob,
             waveform_eob,
-        ) = self.effective_one_body_waveform(params)
+        ) = self.effective_one_body_waveform(params, frequencies)
 
         amplitude_eob, phase_eob = phase_unwrapping(waveform_eob)
 
@@ -272,7 +285,9 @@ class BarePostNewtonianGenerator(WaveformGenerator):
 
         return phase - phase[0]
 
-    def effective_one_body_waveform(self, params: "WaveformParameters"):
+    def effective_one_body_waveform(
+        self, params: "WaveformParameters", frequencies: Optional[list[float]] = None
+    ):
         raise NotImplementedError(
             "This generator does not include the possibility "
             "to generate effective one body waveforms"
@@ -287,7 +302,7 @@ class TEOBResumSGenerator(BarePostNewtonianGenerator):
         self.eobrun_callable = eobrun_callable
 
     def effective_one_body_waveform(
-        self, params: "WaveformParameters"
+        self, params: "WaveformParameters", frequencies: Optional[list[float]] = None
     ) -> tuple[np.ndarray, np.ndarray]:
         r"""Generate an EOB waveform with TEOB.
 
@@ -317,11 +332,31 @@ class TEOBResumSGenerator(BarePostNewtonianGenerator):
         new_f0 = f_0 - delta_f * n_additional
         par_dict["initial_frequency"] = new_f0
 
+        to_slice = (
+            slice(-len(frequencies), None)
+            if frequencies is not None
+            else slice(n_additional, None)
+        )
+
+        if frequencies is not None:
+            frequencies = list(
+                np.insert(
+                    frequencies,
+                    0,
+                    np.arange(f_0 - delta_f * n_additional, f_0, step=delta_f),
+                )
+            )
+            par_dict.pop("df")
+            par_dict["interp_freqs"] = "yes"
+            par_dict["freqs"] = frequencies
+
         f_spa, rhpf, ihpf, _, _ = self.eobrun_callable(par_dict)
 
-        f_spa = f_spa[n_additional:]
+        f_spa = f_spa[to_slice]
+        # f_spa = f_spa
 
-        waveform = (rhpf - 1j * ihpf)[n_additional:]
+        waveform = (rhpf - 1j * ihpf)[to_slice]
+        # waveform = rhpf - 1j * ihpf
 
         return (f_spa, waveform)
 
@@ -713,6 +748,7 @@ class Dataset:
     # TODO: the name of this class is misleading,
     as it contains all information contained to generate
     the dataset but not the data itself.
+    (but I cannot think of a better one, maybe DatasetMeta?)
 
     The amplitude residuals are defined as
     :math:`\log(A _{\text{EOB}} / A_{\text{PN}})`,
@@ -745,11 +781,22 @@ class Dataset:
             needs to reference the dataset and therefure must be created after it.
             Defaults to UniformParameterGenerator.
     parameter_generator_kwargs : dict[str, Any]
-
+            Arguments for the creation of the parameter generator.
     seed : int
             Seed for the random number generator used when generating
             waveforms for the training.
             Defaults to 42.
+    multibanding: bool
+            Whether to use multibanding for the default frequency array.
+            If True, the frequency array is computed according to
+            :func:`reduced_frequency_array`;
+            if False, the frequency array is the "default FFT" one
+            with spacing :attr:`delta_f_hz`.
+            Defaults to False.
+    f_pivot_hz: float
+            Pivot frequency for the multibanding in Hz, only used if
+            :attr:`multibanding` is True.
+            Defaults to 20.
 
     Examples
     --------
@@ -764,10 +811,6 @@ class Dataset:
             Defaults to 2.8; this does not typically need to be changed.
     """
 
-    # TODO
-    # saving to file to be managed with https://docs.h5py.org/en/stable/quick.html
-
-    # total mass of the binary, in solar masses
     total_mass: float = 2.8
 
     def __init__(
@@ -779,6 +822,8 @@ class Dataset:
         parameter_generator_class: Type[ParameterGenerator] = UniformParameterGenerator,
         parameter_generator_kwargs: Optional[dict[str, Any]] = None,
         seed: int = 42,
+        multibanding: bool = True,
+        f_pivot_hz: float = 20.0,
     ):
 
         self.initial_frequency_hz = initial_frequency_hz
@@ -795,6 +840,9 @@ class Dataset:
         self.residuals_amp: list[np.ndarray] = []
         self.residuals_phi: list[np.ndarray] = []
 
+        self.multibanding = multibanding
+        self.f_pivot_hz = f_pivot_hz
+
     def __repr__(self) -> str:
         return (
             f"{self.__class__.__name__}("
@@ -806,27 +854,46 @@ class Dataset:
 
     @property
     def waveform_length(self) -> int:
-        return (
-            int((self.srate_hz / 2 - self.initial_frequency_hz) / self.delta_f_hz) + 1
-        )
+        if self.multibanding:
+            return len(self.frequencies)
+        else:
+            return (
+                int((self.srate_hz / 2 - self.initial_frequency_hz) / self.delta_f_hz)
+                + 1
+            )
 
-    @cached_property
+    @property
     def frequencies(self) -> np.ndarray:
         """Frequency array corresponding to this dataset,
         in natural units.
         """
+        return self._frequencies()
+
+    @lru_cache(maxsize=1)
+    def _frequencies(self):
         return self.hz_to_natural_units(self.frequencies_hz)
 
-    @cached_property
+    @property
     def frequencies_hz(self) -> np.ndarray:
         """Frequency array corresponding to this dataset,
         in Hz.
         """
-        return np.arange(
-            self.initial_frequency_hz,
-            self.srate_hz / 2 + self.delta_f_hz,
-            self.delta_f_hz,
-        )
+        return self._frequencies_hz()
+
+    @lru_cache(maxsize=1)
+    def _frequencies_hz(self):
+        if self.multibanding:
+            return reduced_frequency_array(
+                self.initial_frequency_hz,
+                self.srate_hz / 2,
+                self.f_pivot_hz,
+            )
+        else:
+            return np.arange(
+                self.initial_frequency_hz,
+                self.srate_hz / 2 + self.delta_f_hz,
+                self.delta_f_hz,
+            )
 
     def optimal_df_hz(
         self, power_of_two: bool = True, margin_percent: float = 8.0
@@ -880,9 +947,27 @@ class Dataset:
 
     @property
     def mass_sum_seconds(self) -> float:
+        """Reference total mass expressed in seconds, :math:`GM / c^3`.
+
+        Returns
+        -------
+        float
+        """
         return self.total_mass * SUN_MASS_SECONDS
 
     def hz_to_natural_units(self, frequency_hz: Union[float, np.ndarray]):
+        """Utility function: convert Hz to natural units,
+        using the reference total mass of the dataset.
+
+        Parameters
+        ----------
+        frequency_hz : Union[float, np.ndarray]
+
+        Returns
+        -------
+        frequency_nu
+            Frequency in natural units.
+        """
         return frequency_hz * self.mass_sum_seconds
 
     def taylor_f2_prefactor(self, eta: float) -> float:
@@ -916,6 +1001,18 @@ class Dataset:
     def make_parameter_generator(
         self, seed: Optional[int] = None
     ) -> ParameterGenerator:
+        """Make a new parameter generator,
+        of the type determined by :attr:`parameter_generator_class`.
+
+        Parameters
+        ----------
+        seed : int, optional
+            Seed for the RNG inside the parameter generator, by default None
+
+        Returns
+        -------
+        ParameterGenerators
+        """
         if seed is None:
             seed = self.seed_sequence.integers(low=0, high=2 ** 63 - 1)
 
@@ -978,7 +1075,9 @@ class Dataset:
             (
                 amp_residuals[i],
                 phi_residuals[i],
-            ) = self.waveform_generator.generate_residuals(params, downsampling_indices)
+            ) = self.waveform_generator.generate_residuals(
+                params, list(self.frequencies), downsampling_indices
+            )
 
             parameter_array[i] = params.array
 
@@ -1090,7 +1189,9 @@ class Dataset:
         phis = []
 
         for par in tqdm(waveform_param_list, unit="waveforms"):
-            _, cartesian_wf = self.waveform_generator.effective_one_body_waveform(par)
+            _, cartesian_wf = self.waveform_generator.effective_one_body_waveform(
+                par, list(self.frequencies)
+            )
             amp, phi = phase_unwrapping(cartesian_wf)
             amps.append(amp[amp_indices])
             phis.append(phi[phi_indices])
