@@ -24,6 +24,7 @@ from .data_management import (
 from .dataset_generation import (
     BarePostNewtonianGenerator,
     Dataset,
+    ParameterGenerator,
     ParameterSet,
     TEOBResumSGenerator,
     WaveformGenerator,
@@ -122,15 +123,6 @@ class ParametersWithExtrinsic:
         The parameters are all converted to natural units.
         """
         base_dict = self.intrinsic(dataset).teobresums
-        frequency_rescaling = dataset.total_mass / self.total_mass
-
-        assert isinstance(base_dict["initial_frequency"], float)
-        assert isinstance(base_dict["srate_interp"], float)
-        assert isinstance(base_dict["df"], float)
-
-        base_dict["initial_frequency"] *= frequency_rescaling
-        base_dict["srate_interp"] *= frequency_rescaling
-        base_dict["df"] *= frequency_rescaling
 
         return {
             **base_dict,
@@ -179,6 +171,25 @@ class Model:
     nn_kind : Type[NeuralNetwork]
             Neural network implementation to use,
             defaults to :class:`SklearnNetwork`.
+    parameter_generator_kwargs: dict[str, tuple[float, float]], optional
+            Dictionary of keyword arguments to be used when initializing
+            a :class:`ParameterGenerator`.
+            It should be a map of strings to tuples, in the form
+            `{'q_range': (1., 3.)}`.
+
+            Options include
+            ``q_range``,
+            ``lambda1_range``,
+            ``lambda2_range``,
+            ``chi1_range``,
+            ``chi2_range``.
+    mass_range: tuple[float, float]
+            Range of total masses :math:`M` (in solar masses)
+            that the model should be able to reconstruct.
+            The reconstruction works with the frequency expressed as :math:`Mf`,
+            but changing the total mass changes the range of relevant frequencies.
+
+            Defaults to (2.8, 2.8).
     """
 
     def __init__(
@@ -187,9 +198,12 @@ class Model:
         initial_frequency_hz: float = 20.0,
         srate_hz: float = 4096.0,
         pca_components_number: int = 30,
+        multibanding: bool = True,
         waveform_generator: Optional[WaveformGenerator] = None,
         downsampling_training: Optional[DownsamplingTraining] = None,
         nn_kind: Type[NeuralNetwork] = SklearnNetwork,
+        parameter_generator_kwargs: Optional[dict[str, tuple[float, float]]] = None,
+        mass_range: tuple[float, float] = (2.8, 2.8),
     ):
 
         self.filename = filename
@@ -206,8 +220,16 @@ class Model:
         else:
             self.waveform_generator = waveform_generator
 
+        effective_initial_frequency, effective_srate = expand_frequency_range(
+            initial_frequency_hz, srate_hz, mass_range, Dataset.total_mass
+        )
+
         self.dataset = Dataset(
-            initial_frequency_hz, srate_hz, waveform_generator=self.waveform_generator
+            effective_initial_frequency,
+            effective_srate,
+            waveform_generator=self.waveform_generator,
+            multibanding=multibanding,
+            parameter_generator_kwargs=parameter_generator_kwargs,
         )
 
         if downsampling_training is None:
@@ -566,16 +588,36 @@ class Model:
         )
 
     def predict(self, frequencies: np.ndarray, params: ParametersWithExtrinsic):
-        """Calculate the waveforms in the plus and cross polarizations,
+        r"""Calculate the waveforms in the plus and cross polarizations,
         accounting for extrinsic parameters
 
         Parameters
         ----------
         frequencies : np.ndarray
                 Frequencies where to compute the waveform, in Hz.
+
+                These should always be within the range in which the
+                model has been trained, and be careful!
+                The model is always trained with a specific initial frequency
+                :math:`f_0`, and a final frequency :math:`f_1`,
+                and it is trained to reconstruct the dependence
+                of the waveform on :math:`M_0 f`, where :math:`M_0` is
+                some standard mass, typically :math:`2.8M_{\odot}`.
+
+                Now, this means that the model can only predict in the range
+                :math:`M_0 f_0 \leq M f \leq M_0 f_1`;
+                when :math:`M` differs significantly from :math:`M_0`
+                this will be quite a different range from :math:`[f_0, f_1]`.
+
+
         params : ParametersWithExtrinsic
                 Parameters for the waveform, both intrinsic and extrinsic.
 
+        Raises
+        ------
+        AssertionError
+                When the frequencies given are either too high or too low.
+
         Returns
         -------
         hp, hc (complex np.ndarray)
@@ -586,6 +628,13 @@ class Model:
         assert self.downsampling_indices is not None
         assert self.nn is not None
 
+        rescaled_frequencies = frequencies * (
+            params.total_mass / self.dataset.total_mass
+        )
+
+        assert min(rescaled_frequencies) >= self.dataset.initial_frequency_hz
+        assert max(rescaled_frequencies) <= self.dataset.srate_hz / 2.0
+
         intrinsic_params = params.intrinsic(self.dataset)
 
         residuals = self.predict_residuals_bulk(
@@ -606,13 +655,14 @@ class Model:
 
         amp = self.downsampling_training.resample(
             self.dataset.frequencies_hz[self.downsampling_indices.amplitude_indices],
-            frequencies,
+            rescaled_frequencies,
             amp_ds,
         )
+
         phi = (
             self.downsampling_training.resample(
                 self.dataset.frequencies_hz[self.downsampling_indices.phase_indices],
-                frequencies,
+                rescaled_frequencies,
                 phi_ds,
             )
             + params.reference_phase
@@ -693,27 +743,66 @@ class ModelPredictingInverted(Model):
         return compute_polarizations(cartesian_waveform, pre_plus, pre_cross)
 
 
-###### TEMP FUNCTIONS
-
-
 @njit
-def combine_amp_phase(amp, phase):
+def combine_amp_phase(amp: np.ndarray, phase: np.ndarray) -> np.ndarray:
+    """Combine amplitude and phase arrays into a Cartesian waveform,
+    according to
+    :math:`h = A e^{i \phi}`.
+
+    This function is separated out just so that it can be decorated with ``@njit``.
+
+    Parameters
+    ----------
+    amp : np.ndarray
+    phase : np.ndarray
+
+    Returns
+    -------
+    np.ndarray
+    """
     return amp * np.exp(1j * phase)
 
 
 @njit
-def combine_residuals_amp(amp, amp_pn):
+def combine_residuals_amp(amp: np.ndarray, amp_pn: np.ndarray) -> np.ndarray:
+    r"""Combine amplitude residuals with their Post-Newtonian counterparts,
+    according to
+    :math:`A = A_{PN} e^{\Delta A}`.
+
+    This function is separated out just so that it can be decorated with ``@njit``.
+
+    Parameters
+    ----------
+    amp : np.ndarray
+    amp_pn : np.ndarray
+
+    Returns
+    -------
+    np.ndarray
+    """
     return amp_pn * np.exp(amp)
 
 
 @njit
-def combine_residuals_phi(phi, phi_pn):
+def combine_residuals_phi(phi: np.ndarray, phi_pn: np.ndarray) -> np.ndarray:
+    """Combine amplitude residuals with their Post-Newtonian counterparts,
+    according to
+    :math:`\phi = \phi_{PN} + \Delta \phi`.
+
+    This function is separated out just so that it can be decorated with ``@njit``.
+
+    Parameters
+    ----------
+    phi : np.ndarray
+    phi_pn : np.ndarray
+
+    Returns
+    -------
+    np.ndarray
+    """
     return phi_pn + phi
 
 
-###### END TEMP FUNCTIONS
-
-
 @njit
 def compute_polarizations(
     waveform: np.ndarray,
@@ -848,3 +937,48 @@ def compute_cartesian_waveform(
     )
 
     return amp * np.exp(1j * phi)
+
+
+def expand_frequency_range(
+    initial_frequency: float,
+    final_frequency: float,
+    mass_range: tuple[float, float],
+    reference_mass: float,
+) -> tuple[float, float]:
+    """Widen the frequency range to account for the
+    different masses the user requires.
+
+    Parameters
+    ----------
+    initial_frequency : float
+        Lower bound for the frequency.
+        Typically in Hz, but this function just requires it
+        to be consistent with the other parameters.
+    final_frequency : float
+        Upper bound for the frequency.
+        It can also be given as the time-domain
+        signal rate :math:`r = 1 / \Delta t`, which is
+        twice che maximum frequency because of the Nyquist bound.
+
+        Since all this function does is multiply it by a certain factor,
+        the formulations can be exchanged.
+    mass_range : tuple[float, float]
+        Range of allowed masses, in the same unit as the
+        reference mass (typically, solar masses).
+    reference_mass : float
+        Reference mass the model uses to convert frequencies
+        to the dimensionless :math:`Mf`.
+
+    Returns
+    -------
+    tuple[float, float]
+        New lower and upper bounds for the frequency range.
+    """
+
+    m_min, m_max = mass_range
+    assert m_min <= m_max
+
+    return (
+        initial_frequency * (reference_mass / m_max),
+        final_frequency * (reference_mass / m_min),
+    )
